# Copyright (c) 2012, Siu Kwan Lam
# All rights reserved.
#
# Dummy classes for implementating the Pymothoa dialect.
#

class Construct(object):
    def __init__(self, **kwargs):
        '''Dummy do-nothing implementation for Python code.
        '''
        pass

class var(Construct):
    pass
